from __future__ import annotations

import logging
from typing import Union

from cognite.client._api.assets import AssetsAPI
from cognite.client._api.events import EventsAPI
from cognite.client._api.files import FilesAPI
from cognite.client._api.labels import LabelsAPI
from cognite.client._api.relationships import RelationshipsAPI
from cognite.client._api.sequences import SequencesAPI
from cognite.client.data_classes import Asset, Event, FileMetadata, LabelDefinition, Relationship, Sequence, TimeSeries
from cognite.client.exceptions import CogniteDuplicatedError

logger = logging.getLogger(__name__)


def simple_relationship(
    source: Union[Asset, TimeSeries, FileMetadata, Sequence, Event],
    target: Union[Asset, TimeSeries, FileMetadata, Sequence, Event],
    label_external_id: str,
) -> Relationship:
    """Simplifies Cognite Python SDK creation of Relationships."""
    external_id = f"{source.external_id}.{target.external_id}"
    source_type = "file" if isinstance(source, FileMetadata) else source.__class__.__name__
    target_type = "file" if isinstance(target, FileMetadata) else target.__class__.__name__

    return Relationship(
        external_id=external_id,
        source_type=source_type,
        target_type=target_type,
        source_external_id=source.external_id,
        target_external_id=target.external_id,
        labels=[label_external_id],
    )


CogniteAPI = Union[
    AssetsAPI,
    EventsAPI,
    FilesAPI,
    SequencesAPI,
    LabelsAPI,
    RelationshipsAPI,
]


def upsert_cognite_resources(
    api: CogniteAPI,
    resource_type: str,
    resources: list[Asset | Event | Sequence | LabelDefinition | Relationship],
) -> None:
    try:
        api.create(resources)
    except CogniteDuplicatedError as e:
        duplicated_ids = [ids["externalId"] for ids in e.duplicated]

        # Since the non-existing resource are not created when create fails, we need to create them in a separate call.
        if failed_resources := [r for r in e.failed if r.external_id not in duplicated_ids]:
            api.create(failed_resources)

        logger.debug(f"The following {resource_type} already exists in CDF: {duplicated_ids}")
        resources_to_update = [r for r in resources if r.external_id in duplicated_ids]
        if hasattr(api, "update"):
            api.update(resources_to_update)
        elif resource_type == "label_definitions":
            # For labels, the update method is not implemented. Hence, we need to delete and recreate.
            logger.debug("Deleting existing label definitions")
            api.delete(external_id=duplicated_ids)
            logger.debug("Creating new label definitions")
            api.create(resources_to_update)



from typing import Optional

from cognite.client import ClientConfig, CogniteClient

from cognite.powerops.client.asset_apis import (
    BenchmarkingConfigurationsAPI,
    BidConfigurationsAPI,
    GeneratorsAPI,
    MarketConfigurationsAPI,
    PlantAPI,
    PriceAreasAPI,
    ReservoirsAPI,
    RKOMBidCombinationConfiguration,
    RKOMBidConfigurationsAPI,
    WatercourseAPI,
)
from cognite.powerops.client.config_client import ConfigurationClient
from cognite.powerops.client.dm.client import get_power_ops_dm_client
from cognite.powerops.client.dm_apis import CaseAPI, CommandsAPI, MappingAPI, ScenarioAPI, TransformationAPI
from cognite.powerops.client.shop_api import ShopAPI


class ConfigurationsClient:
    def __init__(self, read_dataset: str, write_dataset: str, client: CogniteClient):
        self.bids = BidConfigurationsAPI(client, read_dataset, write_dataset)
        self.rkom_bids = RKOMBidConfigurationsAPI(client, read_dataset, write_dataset)
        self.bechmarkings = BenchmarkingConfigurationsAPI(client, read_dataset, write_dataset)
        self.markets = MarketConfigurationsAPI(client, read_dataset, write_dataset)
        self.rkom_bid_combinations = RKOMBidCombinationConfiguration(client, read_dataset, write_dataset)


class PowerOpsClient:
    def __init__(self, read_dataset: str, write_dataset: str, config: Optional[ClientConfig] = None):
        self.dm = get_power_ops_dm_client(config=config)
        self.cdf = self.dm._client

        self.configurations = ConfigurationClient()

        self.shop = ShopAPI(po_client=self)

        self.configurations = ConfigurationsClient(read_dataset, write_dataset, self.cdf)
        self.generators = GeneratorsAPI(self.cdf, read_dataset, write_dataset)
        self.plants = PlantAPI(self.cdf, read_dataset, write_dataset)
        self.price_areas = PriceAreasAPI(self.cdf, read_dataset, write_dataset)
        self.reservoirs = ReservoirsAPI(self.cdf, read_dataset, write_dataset)
        self.watercourses = WatercourseAPI(self.cdf, read_dataset, write_dataset)

        self.cases = CaseAPI(self.dm)
        self.commands = CommandsAPI(self.dm)
        self.scenarios = ScenarioAPI(self.dm)
        self.mappings = MappingAPI(self.dm)
        self.transformations = TransformationAPI(self.dm)



from __future__ import annotations

import logging
import tempfile
from typing import Any

import yaml

logger = logging.getLogger(__name__)


class Case:
    r"""
    Wrapper around YAML file for SHOP, describing a case.

    Main features:
     * Values can be retrieved or set / changed like with a dict.
        * Dot syntax is also supported.
     * Can hold references to additional files ("extra files").
        * These are just paths to files, their content is not accessed here.
     * Loading from string or file (with `Case.load_yaml(path_to_yaml)`).
     * Saving to yaml (with `case.save_yaml(path_to_file)`).

    Examples:
      * load a case from string
          >>> case = Case('''
          ... foo:
          ...   bar1: 11
          ...   bar2: 22
          ... ''')
          >>> case.data
          {'foo': {'bar1': 11, 'bar2': 22}}

      * edit a value and show updated data
          >>> case["foo"]["bar2"] = 202
          >>> case.yaml
          'foo:\n  bar1: 11\n  bar2: 202\n'

      * dot-notation, just for convenience
          >>> case["foo.bar1"] = 101
          >>> case["foo.bar1"]
          101

      * load from and save to a file:
          case = Case.load_yaml("path/to/my_case.yaml")
          case[...] = ...  # edit data
          case.save_yaml("path/to/same_or_different.yaml")
    """

    def __init__(self, data: str = "") -> None:
        self._data = {}
        self._extra_files = []

        yaml_docs = list(yaml.safe_load_all(data))
        if yaml_docs:
            self._data = yaml_docs[0]
            for yaml_doc in yaml_docs[1:]:
                tmp_file = tempfile.NamedTemporaryFile(
                    mode="w",
                    delete=False,
                    prefix="powerops-sdk-tmp-",
                    suffix=".yaml",
                )
                tmp_file.write(yaml.dump(yaml_doc))
                tmp_file.close()
                self.add_extra_file(tmp_file.name)

    @classmethod
    def load_yaml(cls, yaml_path: str) -> Case:
        logger.info(f"loading case file: {yaml_path}")
        with open(yaml_path, "r") as yaml_file:
            return cls(yaml_file.read())

    def add_extra_file(self, file_path: str) -> None:
        logger.info(f"adding extra file: {file_path}")
        self._extra_files.append(file_path)

    @property
    def extra_files(self):
        return self._extra_files.copy()

    @property
    def data(self) -> dict:
        return self._data.copy()

    def __getitem__(self, path: str) -> Any:
        """Get item from `self._data` dict, supporting "." as separator"""
        if path in self._data:  # edge case: key actually contains a dot
            return self._data[path]
        part_data = self.data
        for part in path.split("."):
            if isinstance(part_data, list):
                part = int(part)
            part_data = part_data[part]
        return part_data

    def __setitem__(self, path: str, value: Any) -> None:
        """Set item from `self._data` dict, supporting "." as separator"""
        if path in self._data:  # edge case: key actually contains a dot
            self._data[path] = value
        else:
            part_data = self.data
            parts = path.split(".")
            while parts:
                part = parts.pop(0)
                if isinstance(part_data, list):
                    part = int(part)
                if parts:
                    part_data = part_data[part]
                else:
                    # last loop:
                    part_data[part] = value

    @property
    def yaml(self) -> str:
        return yaml.dump(self._data, sort_keys=False)

    def save_yaml(self, path: str) -> None:
        logger.info(f"Saving case file to: {path}")
        with open(path, "w") as output_file:
            output_file.write(yaml.dump(self._data, sort_keys=False))
